import numpy as np
from typing import Tuple

import torch


class LiDARFilter(object):
    def __init__(self,
                 close_radius=1.0,
                 x_range='(-50.0, 50.0)',
                 y_range='(-50.0, 50.0)',
                 z_range='(-5, 20)',
                 ):
        self.radius = close_radius
        self.x_range = eval(x_range)
        self.y_range = eval(y_range)
        self.z_range = eval(z_range)

    def _remove_close(self, points, radius=1.0):
        """Removes point too close within a certain radius from origin.

        Args:
            points (np.ndarray | :obj:`BasePoints`): Sweep points.
            radius (float, optional): Radius below which points are removed.
                Defaults to 1.0.

        Returns:
            np.ndarray: Points after removing.
        """
        x_filt = torch.abs(points[:, 0]) < radius
        y_filt = torch.abs(points[:, 1]) < radius
        not_close = torch.logical_not(torch.logical_and(x_filt, y_filt))
        return points[not_close]

    def range_filter(
        self,
        points,
        xrange: Tuple[float, float] = (-np.inf, np.inf),
        yrange: Tuple[float, float] = (-np.inf, np.inf),
        zrange: Tuple[float, float] = (-np.inf, np.inf),
    ) -> None:
        """
        Restricts points to specified ranges.
        :param xrange: (xmin, xmax).
        :param yrange: (ymin, ymax).
        :param zrange: (zmin, zmax).
        """
        # Figure out which points to keep.
        keep_x = torch.logical_and(xrange[0] <= points[:, 0], points[:, 0] <= xrange[1])
        keep_y = torch.logical_and(yrange[0] <= points[:, 1], points[:, 1] <= yrange[1])
        keep_z = torch.logical_and(zrange[0] <= points[:, 2], points[:, 2] <= zrange[1])
        keep = torch.logical_and(keep_x, torch.logical_and(keep_y, keep_z))
        return points[keep]


    def __call__(self, features, targets):
        """Call function to load multi-sweep point clouds from files.

        Args:
            results (dict): Result dict containing multi-sweep point cloud
                filenames.

        Returns:
            dict: The result dict containing the multi-sweep points data.
                Added key and value are described below.

                - points (np.ndarray | :obj:`BasePoints`): Multi-sweep point
                    cloud arrays.
        """
        points = torch.cat(features['lidars_warped'], 0)
        points = self._remove_close(points, self.radius)
        points = self.range_filter(points, self.x_range, self.y_range, self.z_range)
        features['lidar'] = points
        return features, targets
