from typing import Dict, Tuple, List

import numpy as np
import pytorch_lightning as pl
import torch
from nuplan.planning.simulation.trajectory.trajectory_sampling import TrajectorySampling
from torch import Tensor

from navsim.agents.abstract_agent import AbstractAgent
from navsim.agents.transfuser.transfuser_agent import TransfuserAgent
from navsim.agents.hydra.consistency_agent import HydraAgentPEcon
from navsim.agents.vadv2.vadv2_agent import Vadv2Agent
from navsim.common.dataclasses import Trajectory
import time

class AgentLightningModule(pl.LightningModule):
    def __init__(
            self,
            agent: AbstractAgent,
    ):
        super().__init__()
        self.agent = agent
        self.prediction_start_time = None
        self.prediction_end_time = None
        self.total_predictions = 0
    def _step(
            self,
            batch: Tuple[Dict[str, Tensor], Dict[str, Tensor], List[str]],
            logging_prefix: str,
    ):
        features, targets, tokens = batch
        if logging_prefix in ['train', 'val'] and isinstance(self.agent, Vadv2Agent):
            prediction = self.agent.forward_train(features, targets['interpolated_traj'])
        else:
            prediction = self.agent.forward(features)

        if isinstance(self.agent, TransfuserAgent) or isinstance(self.agent, Vadv2Agent):
            loss, loss_dict = self.agent.compute_loss(features, targets, prediction)
        else:
            loss, loss_dict = self.agent.compute_loss(features, targets, prediction, tokens)
        # if 'consistency_loss' in loss_dict and self.current_epoch <= 5:
        #     loss = loss - loss_dict['consistency_loss']
        #     loss_dict.pop('consistency_loss', None)

        for k, v in loss_dict.items():
            self.log(f"{logging_prefix}/{k}", v, on_step=True, on_epoch=True, prog_bar=True, sync_dist=True)
        self.log(f"{logging_prefix}/loss", loss, on_step=True, on_epoch=True, prog_bar=True, sync_dist=True)
        return loss

    def training_step(
            self,
            batch: Tuple[Dict[str, Tensor], Dict[str, Tensor]],
            batch_idx: int
    ):
        return self._step(batch, "train")

    def validation_step(
            self,
            batch: Tuple[Dict[str, Tensor], Dict[str, Tensor]],
            batch_idx: int
    ):
        return self._step(batch, "val")

    def on_predict_start(self):
        # 开始预测时记录时间
        self.prediction_start_time = time.time()
        self.total_predictions = 0  # 重置预测计数器

    def on_predict_end(self):
        # 结束预测时记录时间
        self.prediction_end_time = time.time()
        prediction_time = self.prediction_end_time - self.prediction_start_time
        prediction_speed = self.total_predictions / prediction_time  # 计算推理速度 (it/s)
        print(f"Prediction speed: {prediction_speed:.2f} iterations per second")

    def validation_step_end(self, outputs):
        # 每次 step 结束时统计 batch 数
        self.total_iterations += 1

    def configure_optimizers(self):
        return self.agent.get_optimizers()

    def predict_step(
            self,
            batch: Tuple[Dict[str, Tensor], Dict[str, Tensor]],
            batch_idx: int
    ):
        self.total_predictions += 1
        if isinstance(self.agent, TransfuserAgent) or isinstance(self.agent, Vadv2Agent):
            return self.predict_step_transfuser(batch, batch_idx)
        else:
            return self.predict_step_hydra(batch, batch_idx)

    # ablate overall pdm score
    def predict_step_transfuser(
            self,
            batch: Tuple[Dict[str, Tensor], Dict[str, Tensor]],
            batch_idx: int
    ):
        features, targets, tokens = batch
        self.agent.eval()
        with torch.no_grad():
            predictions = self.agent.forward(features)
            poses = predictions["trajectory"].cpu().numpy()
            if 'trajectory_pre' in predictions:
                poses_pre = predictions["trajectory_pre"].cpu().numpy()
            else:
                poses_pre = poses

        if poses.shape[1] == 40:
            interval_length = 0.1
        else:
            interval_length = 0.5

        return {token: {
            'trajectory': Trajectory(pose, TrajectorySampling(time_horizon=4, interval_length=interval_length)),
            'trajectory_pre': Trajectory(pose_pre, TrajectorySampling(time_horizon=4, interval_length=interval_length)),

        } for pose, pose_pre, token in zip(poses, poses_pre, tokens)}

    # hydra-pdm
    def predict_step_hydra(
            self,
            batch: Tuple[Dict[str, Tensor], Dict[str, Tensor]],
            batch_idx: int
    ):
        features, targets, tokens = batch
        self.agent.eval()
        with torch.no_grad():
            predictions = self.agent.forward(features)
            poses = predictions["trajectory"].cpu().numpy()
            if 'trajectory_pre' in predictions:
                poses_pre = predictions['trajectory_pre'].cpu().numpy()
            else:
                poses_pre = poses

            imis = predictions["imi"].softmax(-1).log().cpu().numpy()
            nocs = predictions["noc"].log().cpu().numpy()
            das = predictions["da"].log().cpu().numpy()
            ttcs = predictions["ttc"].log().cpu().numpy()
            if 'comfort' in predictions:
                comforts = predictions["comfort"].log().cpu().numpy()
            else:
                comforts = np.zeros(len(tokens))
            if 'lk' in predictions:
                lks = predictions["lk"].log().cpu().numpy()
            else:
                lks = np.zeros(len(tokens))
            if 'tl' in predictions:
                tls = predictions["tl"].log().cpu().numpy()
            else:
                tls = np.zeros(len(tokens))
            if 'dr' in predictions:
                drs = predictions["dr"].log().cpu().numpy()
            else:
                drs = np.zeros(len(tokens))
            if 'progress' in predictions:
                progresses = predictions["progress"].log().cpu().numpy()
            else:
                progresses = [None for _ in range(len(tokens))]
        if poses.shape[1] == 40:
            interval_length = 0.1
        else:
            interval_length = 0.5

        return {token: {
            'trajectory': Trajectory(pose, TrajectorySampling(time_horizon=4, interval_length=interval_length)),
            'trajectory_pre': Trajectory(pose_pre, TrajectorySampling(time_horizon=4, interval_length=interval_length)),
            'imi': imi,
            'noc': noc,
            'da': da,
            'ttc': ttc,
            'comfort': comfort,
            'progress': progress,
            'lk': lk,
            'tl': tl,
            'dr': dr,
        } for pose, pose_pre, imi, noc, da, ttc, comfort, progress, lk, tl, dr, token in
            zip(poses, poses_pre, imis, nocs, das, ttcs, comforts, progresses,
                lks, tls, drs, tokens)}


    # ablate post-processing vadv2
    # def predict_step(
    #         self,
    #         batch: Tuple[Dict[str, Tensor], Dict[str, Tensor]],
    #         batch_idx: int
    # ):
    #     features, _, tokens = batch
    #     self.agent.eval()
    #     K = 100
    #     # N_VOCAB, 40, 3
    #     vocab = self.agent.vadv2_model._trajectory_head.vocab
    #     with torch.no_grad():
    #         predictions = self.agent.forward(features)
    #         # poses = predictions["trajectory"].cpu().numpy()
    #         # B, N_VOCAB
    #         imi_score = predictions["trajectory_distribution"].softmax(-1).log()
    #         # B, K
    #         topk_scores, topk_inds = imi_score.topk(K, -1)
    #         # B, K, 40->20, 3->2
    #         topk_trajs = vocab[topk_inds][:, :, :20, :2]
    #
    #         # B, 30, 5 (x,y,h,l,w)
    #         agents = predictions["agent_states"].cpu().numpy()
    #
    #         # B, 7, H=128, W=256
    #         map = predictions["bev_semantic_map"].softmax(1).log().cpu().numpy()
    #         B, _, H, W = map.shape
    #         post_scores = topk_scores.clone()
    #
    #         # normalize trajs
    #         topk_trajs[..., 0] = topk_trajs[..., 0] / 32
    #         topk_trajs[..., 1] = topk_trajs[..., 1] / 32
    #
    #         # B, H, W
    #         good_locs = map[:, 1:2]
    #         bad_locs = map[:, 2:3]
    #         post_scores += F.grid_sample(good_locs, topk_trajs, mode='nearest').sum((-1,)).squeeze(1)
    #         post_scores -= F.grid_sample(bad_locs, topk_trajs, mode='nearest').sum((-1,)).squeeze(1)
    #
    #         post_ind = post_scores.argmax(-1)
    #         poses = vocab[topk_inds[post_ind]].cpu().numpy()
    #
    #     if poses.shape[1] == 40:
    #         interval_length = 0.1
    #     else:
    #         interval_length = 0.5
    #
    #     return {token: {
    #         'trajectory': Trajectory(pose, TrajectorySampling(time_horizon=4, interval_length=interval_length)),
    #
    #     } for pose, token in zip(poses, tokens)}

